#!/usr/bin/env perl
# Copyright (C) 2011 by Tomasz Konojacki
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Setup stuff...
package Camelwiki;

our $VERSION = '1.2.5';

use strict;
use warnings;
use utf8;
use open qw(:std :utf8);

# Comment out following line if you are debugging.
no warnings qw(once);

################################################################################
# CAMELWIKI CONFIGURATION                                                      #
################################################################################

# Database type, available choices:
#   'mysql' - MySQL.
#   'postgres' - PostgreSQL.
#   'sqlite' - SQLite

my $database_type = '';

# Database name, edit only if you are using MySQL or PostgreSQL database.
# Examples:
#   'foobar'
#   'example'

my $database_name = '';

# Database username, edit only if you are using MySQL or PostgreSQL database.
# Examples:
#   'username'
#   'foobar'

my $database_username = '';

# Database password, edit only if you are using MySQL or PostgreSQL database.
# Examples:
#   'password123'
#   'secure_password'

my $database_password = '';

# Database host, edit only if you are using MySQL or PostgreSQL database.
# Examples:
#   'localhost'
#   '192.168.1.1'
#   'foobar.com'

my $database_host = '';

# Database port number, edit only if you are using MySQL or PostgreSQL database.
# Examples:
#   5432
#   3306

my $database_port = 1234;

# Database filename, edit only if you are using SQLite database.
# Examples:
#   'foobar.db'
#   '/path/to/database.db'

my $database_file = '';

# Name of wiki.
# Examples:
#   'Foobar Wiki'
#   'blah blah'

my $wiki_name = 'I forgot to configure CamelWiki';

# Wiki logo.
# Examples:
#   'C:\path\to\logo\image.png'
#   '/unix/path/to/logo/image.jpg'

my $wiki_logo = 'logo.png';

# Wiki CSS.
# Examples:
#   'style.css'
#   'http://www.server.com/style.css'

my $wiki_css = 'style.css';

# Enable table of contents?
#   1 - yes
#   0 - no

my $enable_table_of_contents = 0;

# Default page.
# Examples:
#   'HomePage'
#   'foobar'

my $default_page = 'HomePage';

# Enable ReCAPTCHA?
# Note that if reCAPTCHA is enabled you need Captcha::reCAPTCHA module. If
# reCaptcha is disabled CamelWiki still has some kind of basic anti-spam
# protection.
#   1 - yes
#   0 - no

my $enable_recaptcha = 0;

# reCAPTCHA public key (to get one register on http://www.recaptcha.net).
# Don't edit if you have disabled reCAPTCHA.

my $recaptcha_public_key = '';

# reCAPTCHA private key (to get one register on http://www.recaptcha.net). Don't
# edit if you have disabled reCAPTCHA.

my $recaptcha_private_key = '';

################################################################################
# END OF CONFIGURATION, DON'T EDIT ANYTHING BELOW THAT LINE!                   #
################################################################################

# Load neccssary modules.
use CGI qw(-utf8 -no_xhtml);
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;
use CGI::Util qw(escape unescape);
use DBI;
use Encode qw(encode decode);

# Load Captcha::reCAPTCHA if reCAPTCHA is enabled.
if ($enable_recaptcha) {
	eval {
		require Captcha::reCAPTCHA;
	} or die $@;
};

{
	# general_url_escape() from Pod::Simple::HTML is broken so it needs to be
	# fixed by following hack.
	package Camelwiki::Pod::Simple::HTML;
	use base qw(Pod::Simple::HTML);
	sub general_url_escape {
		my($self, $string) = @_;
		$string = CGI::Util::escape($string . '');
		return $string;
	}
}

sub connect_to_database {
	# MySQL.
	if ($database_type eq 'mysql') {
		my($dbh, $dsn);
		unless ($database_name) {
			die 'Database name is unspecified!';
		}
		unless ($database_username) {
			die 'Database username is unspecified!';
		}
		unless ($database_password) {
			die 'Database password is unspecified!';
		}
		unless ($database_host) {
			die 'Database host is unspecified!';
		}
		if ($database_port) {
			$dsn = "DBI:mysql:dbname=$database_name;host=$database_host;"
				. "port=$database_port";
		}
		else {
			$dsn = "DBI:mysql:$database_name;host=$database_host;"
		}
		$dbh = DBI->connect(
			$dsn,
			$database_username,
			$database_password,
			{
				PrintError => 0,
				RaiseError => 1
			}
		);
		return $dbh;
	}
	# PostgreSQL.
	elsif ($database_type eq 'postgres') {
		my ($dbh, $dsn);
		unless ($database_name) {
			die 'Database name is unspecified!';
		}
		unless ($database_username) {
			die 'Database username is unspecified!';
		}
		unless ($database_password) {
			die 'Database password is unspecified!';
		}
		unless ($database_host) {
			die 'Database host is unspecified!';
		}
		if ($database_port) {
			$dsn = "DBI:Pg:dbname=$database_name;host=$database_host;"
				. "port=$database_port";
		}
		else {
			$dsn = "DBI:Pg:dbname=$database_name;host=$database_host;"
		}
		$dbh = DBI->connect(
			$dsn,
			$database_username,
			$database_password,
			{
				PrintError => 0,
				RaiseError => 1
			}
		);
		return $dbh;
	}
	# SQLite.
	elsif ($database_type eq 'sqlite') {
		my ($dbh, $dsn);
		unless ($database_file) {
			die 'Specify database filename!';
		}
		$dsn = "dbi:SQLite:dbname=$database_file;";
		$dbh = DBI->connect(
			$dsn,
			'',
			'',
			{
				PrintError => 0,
				RaiseError => 1
			}
		);
		return $dbh;
	}
	# Error, unknown database type!
	else {
		die "Unknown database type: '$database_type'!";
	}
}

sub show_page {
	# This subroutine shows page.
	my ($page, $cgi, $dbh) = @_;
	my (
		$sth, $hashref, $title, $content, $last_modified, $last_modified_by,
		$page_quoted, $pod_html, $pod_html_output, $self_url, $escaped_title,
		$escaped_default_page
	);
	# Escape default page for (for link).
	$escaped_default_page = escape($default_page);
	# Store self url in $self_url.
	$self_url = $cgi->url(-relative => 1) ? $cgi->url(-relative => 1) : '';
	# Encode to UTF-7 and quote $page_qouted (data in database is stored using
	# UTF-7 encoding).
	$page_quoted = $dbh->quote(encode('UTF-7', $page));
	$sth = $dbh->prepare(
		'SELECT title, content, last_modified, last_modified_by FROM pages '
		. "WHERE title = $page_quoted;"
	);
	$sth->execute;
	$hashref = $sth->fetchrow_hashref;
	# Check data received from database
	if (defined $hashref->{title}) {
		$title = decode('UTF-7', $hashref->{title});
	}
	else {
		$title = $page;
	}
	# Escape page title (for link).
	$escaped_title = escape($title);
	if (defined $hashref->{content}) {
		$content = decode('UTF-7', $hashref->{content});
	}
	else {
		# Default page content.
		$content = <<OUTPUT;
=head1 $title

Enter text here.

=cut
OUTPUT
	}
	if (defined $hashref->{last_modified}) {
		$last_modified = $hashref->{last_modified};
	}
	else {
		$last_modified = 'never';
	}
	if (defined $hashref->{last_modified_by}) {
		$last_modified_by = decode('UTF-7', $hashref->{last_modified_by});
	}
	else {
		$last_modified_by = 'Anonymous';
	}
	# Configure Pod::Simple::HTML.
	$pod_html = Camelwiki::Pod::Simple::HTML->new;
	$pod_html->output_string(\$pod_html_output);
	$Camelwiki::Pod::Simple::HTML::Doctype_decl =
		'<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"'
		. "\"http://www.w3.org/TR/html4/loose.dtd\">\n";
	$Camelwiki::Pod::Simple::HTML::Content_decl =
		'<meta http-equiv="Content-Type '
		. 'content="text/html; charset=UTF-8">';
	$pod_html->perldoc_url_prefix("$self_url?");
	$pod_html->html_header_before_title("<html><head><title>$wiki_name - ");
	$pod_html->html_header_after_title(
		'</title><link rel="stylesheet" type="text/css" '
		. "href=\"$wiki_css\"></head><body>"
		. "<a href=\"$self_url?$escaped_default_page\">"
		. "<img src=\"$wiki_logo\" border=\"0\" align=\"right\"</a>"
	);
	$pod_html->html_footer('');
	$pod_html->index($enable_table_of_contents);
	$pod_html->parse_string_document($content);
	# Display data to client.
	print $cgi->header(-charset => 'UTF-8'),
		$pod_html_output,
		# This is *very* dirty hack which allows editing page when POD is empty
		"<p><hr><a href=\"$self_url?page=$escaped_title&action=edit\">"
		. "Edit this page</a> (Last modified: $last_modified by "
		. "$last_modified_by)</p></body></html>";

}

sub edit_page {
	# This subroutine shows page edit form.
	my ($page, $cgi, $dbh) = @_;
	my (
		$sth, $hashref, $title, $content, $last_modified, $last_modified_by,
		$page_quoted, $self_url, $escaped_title, $security_code,
		$security_cookie, $escaped_default_page, $user_name,
		$captcha_html_code, $recaptcha
	);
	# Create Captcha::reCAPTCHA object if reCAPTCHA is enabled.
	if ($enable_recaptcha) {
		eval {
			$recaptcha = Captcha::reCAPTCHA->new;
		} or die $@;
		# Check is reCAPTCHA public key specified
		unless ($recaptcha_public_key) {
			die 'reCAPTCHA public key is unspecified!';
		}
	}
	# Escape default page for (for link).
	$escaped_default_page = escape($default_page);
	# Store self url in $self_url.
	$self_url = $cgi->url(-relative => 1) ? $cgi->url(-relative => 1) : '';
	# Encode to UTF-7 and quote $page_qouted (data in database is stored using
	# UTF-7 encoding).
	$page_quoted = $dbh->quote(encode('UTF-7', $page));
	$sth = $dbh->prepare(
		'SELECT title, content, last_modified, last_modified_by FROM pages '
		. "WHERE title = $page_quoted;"
	);
	$sth->execute;
	$hashref = $sth->fetchrow_hashref;
	# Check data received from database
	if (defined $hashref->{title}) {
		$title = decode('UTF-7', $hashref->{title});
	}
	else {
		$title = $page;
	}
	# Escape page title (for link).
	$escaped_title = escape($title);
	if (defined $hashref->{content}) {
		$content = decode('UTF-7', $hashref->{content});
	}
	else {
		# Default page content.
		$content = <<OUTPUT;
=head1 $title

Enter text here.

=cut
OUTPUT
	}
	if (defined $hashref->{last_modified}) {
		$last_modified = $hashref->{last_modified};
	}
	else {
		$last_modified = 'never';
	}
	if (defined $hashref->{last_modified_by}) {
		$last_modified_by = decode('UTF-7', $hashref->{last_modified_by});
	}
	else {
		$last_modified_by = 'Anonymous';
	}
	# If reCaptcha is enabled there's no need to generate security code.
	unless ($enable_recaptcha) {
		# Generate security code
		$security_code = int(rand(8999)) + 1000;
		# Write security code to cookie
		$security_cookie = CGI::Cookie->new(
			-name  => 'security_code',
			-value => $security_code
		);
	}
	# Read name from cookie or return empty string.
	$user_name = defined $cgi->cookie('username')
		? $cgi->cookie('username')
		: '';
	# Generate reCAPTCHA or plain text verification html code.
	unless ($enable_recaptcha) {
		$captcha_html_code = "Type verification number $security_code:"
			. $cgi->textfield(
				-name    => 'vercode',
				-default => ''
			);
	}
	else {
		eval {
			$captcha_html_code = $recaptcha->get_html($recaptcha_public_key);
		} or die $@;
	}

	print $cgi->header(
			-charset => 'UTF-8',
			-cookie  => [$security_cookie]
		),
		$cgi->start_html(
			-title => "$wiki_name - Editing page $title",
			-style => { 'src' => $wiki_css }
		),
		$cgi->a(
			{
				href => "$self_url?$escaped_default_page"
			},
			$cgi->img({
				align  => 'right',
				border => 0,
				src    => $wiki_logo
			}),
		),
		$cgi->a(
			{
				href => "$self_url?$title"
			},
			$cgi->h1($title)
		),
		$cgi->start_form(-action => $self_url),
		$cgi->hidden(
			-name     => 'action',
			-default  => 'write',
			-override => 1,
		),
		$cgi->hidden(
			-name     => 'page',
			-default  => $title,
			-override => 1,
		),
		$cgi->textarea(
			-name    => 'content',
			-default => $content,
			-rows    => 25,
			-columns => 60
		),
		$cgi->br,
		$cgi->a(
			{
				href => 'http://perldoc.perl.org/perlpod.html'
			},
			'Markup help'
		),
		$cgi->br,
		'Your name (leave it empty if you want to be anonymous):',
		$cgi->textfield(
			-name    => 'name',
			-default => $user_name,
		),
		$cgi->br,
		$cgi->checkbox(
			-name    => 'remember_name',
			-checked => defined $cgi->cookie('remember_name_checked') ? 1 : 0,
			-value   => 'TRUE',
			-label   => 'Check this box if you want to remember your name.'
		),
		$cgi->br,
		$captcha_html_code,
		$cgi->submit(
			-name  => 'submit',
			-value => 'Save'
		),
		$cgi->end_form,
		$cgi->end_html;
}

sub write_page {
	# This subroutine writes modifications to database.
	my ($page, $cgi, $dbh) = @_;
	my (
		$sth, $page_quoted, $hashref, $last_modified_by_quoted, $content_quoted,
		$escaped_title, $self_url, $user_name_cookie,
		$is_remember_name_checked_cookie, $recaptcha, $recaptcha_result
	);
	# Create Captcha::reCAPTCHA object if reCAPTCHA is enabled.
	if ($enable_recaptcha) {
		eval {
			$recaptcha = Captcha::reCAPTCHA->new;
		} or die $@;
		# Check is reCAPTCHA private key specified
		unless ($recaptcha_private_key) {
			die 'reCAPTCHA private key is unspecified!';
		}
	}
	# Store self url in $self_url.
	$self_url = $cgi->url(-relative => 1) ? $cgi->url(-relative => 1) : '';
	# Check params.
	unless (defined $cgi->param('content')) {
		die 'Content unspecified!';
	}
	# Store user name in cookie.
	$user_name_cookie = CGI::Cookie->new(
		-name    => 'username',
		-value   => defined $cgi->param('remember_name')
						? $cgi->param('name')
						: '',
		-expires => '+1M',
	);
	# Store in cookie information if 'remember_name' was checked
	$is_remember_name_checked_cookie = CGI::Cookie->new(
		-name    => 'remember_name_checked',
		-value   => defined $cgi->param('remember_name')
						? 'TRUE'
						: '',
		-expires => '+1M',
	);
	# Quote strings used in database query.
	$page_quoted = $dbh->quote(encode('UTF-7', $page));
	$content_quoted = $dbh->quote(encode('UTF-7', $cgi->param('content')));
	$sth = $dbh->prepare("SELECT title FROM pages WHERE title = $page_quoted");
	$sth->execute;
	$hashref = $sth->fetchrow_hashref;
	unless ($enable_recaptcha) {
		# Check security code.
		if ($cgi->cookie('security_code') != $cgi->param('vercode')) {
				die 'Invalid verification code!';
		}
	}
	else {
		eval {
			$recaptcha_result = $recaptcha->check_answer(
				$recaptcha_private_key,
				$ENV{'REMOTE_ADDR'},
				$cgi->param('recaptcha_challenge_field'),
				$cgi->param('recaptcha_response_field')
			);
			unless ($recaptcha_result->{is_valid}) {
				die 'Invalid verification code!';
			}
		} or die $@;
	}
	if ($cgi->param('name')) {
		$last_modified_by_quoted =
			$dbh->quote(encode('UTF-7', $cgi->param('name')));
	}
	else {
		$last_modified_by_quoted = $dbh->quote(encode('UTF-7', 'Anonymous'));
	}
	unless (defined $hashref->{title}) {
		# Page doesn't exist so script will create new.
		if ($database_type eq 'sqlite') {
			$dbh->do(
				'INSERT INTO pages (title, content, last_modified, '
				. "last_modified_by) VALUES ($page_quoted, $content_quoted, "
				. "CURRENT_TIMESTAMP, $last_modified_by_quoted)"
			);
		}
		else { # PostgreSQL and MySQL
			# MySQL supports LOCALTIMESTAMP() as alias to NOW()
			$dbh->do(
				'INSERT INTO pages (title, content, last_modified, '
				. "last_modified_by) VALUES ($page_quoted, $content_quoted, "
				. "LOCALTIMESTAMP(0), $last_modified_by_quoted)"
			);
		}
	}
	else {
		if ($database_type eq 'sqlite') {
			$dbh->do(
				# Page exist so script will update it.
				"UPDATE pages SET content = $content_quoted, "
				. 'last_modified = CURRENT_TIMESTAMP, '
				. "last_modified_by = $last_modified_by_quoted"
				. "WHERE title = $page_quoted;"
			);
		}
		else { # PostgreSQL and MySQL
			$dbh->do(
				# Page exist so script will update it.
				"UPDATE pages SET content = $content_quoted, "
				. 'last_modified = LOCALTIMESTAMP(0), '
				. "last_modified_by = $last_modified_by_quoted"
				. "WHERE title = $page_quoted;"
			);
		}
	}
	$escaped_title = escape($page);
	print $cgi->redirect(
		-cookie => [$user_name_cookie, $is_remember_name_checked_cookie],
		-uri    => "$self_url?$escaped_title"
	);
}

sub run {
	# This subroutine is equivalent to C main() function.
	# Create CGI object.
	my $cgi = CGI->new;
	# Connect to database.
	my $dbh = connect_to_database();
	# If 'action' param equals 'edit' edit requested page.
	if (
		(defined $cgi->param('action')) && ($cgi->param('action') eq 'edit')
	) {
		unless (defined $cgi->param('page')) {
			die 'Please specify page to edit!'
		}
		edit_page($cgi->param('page'), $cgi, $dbh);
	}
	# If 'action' param equals 'write' write requested page.
	elsif (
		(defined $cgi->param('action')) && ($cgi->param('action') eq 'write')
	) {
		unless (defined $cgi->param('page')) {
			die 'Please specify page to write!'
		}
		# Page title must not contain whitespaces.
		$cgi->param('page') =~ m/^\s*([^\s]*)/;
		write_page($1, $cgi, $dbh);
	}
	# If there is 'page' param show requested page.
	elsif (defined $cgi->param('page')) {
		# Page title must not contain whitespaces.
		$cgi->param('page') =~ m/^\s*([^\s]*)/;
		show_page($1, $cgi, $dbh);
	}
	# If there is no 'page' param but query string exists show requested page.
	elsif ((defined $ENV{QUERY_STRING}) && ($ENV{QUERY_STRING} ne '')) {
		# Page title must not contain whitespaces.
		decode('UTF-8', unescape($ENV{QUERY_STRING})) =~ m/^\s*([^\s]*)/;
		show_page($1, $cgi, $dbh);
	}
	# There is no query string so show home page.
	else {
		# Page title must not contain whitespaces.
		$default_page =~ m/^\s*([^\s]*)/;
		show_page($1, $cgi, $dbh);
	}
}

run();
